import sys
import pygame
import os.path
import pygcurse
import traceback
import inspect
if __name__ == '__main__':
    print("This file is not meant to be executed. Please try main.py.")
    exit()

import gameobjs
import code

class InjectionConsole(code.InteractiveConsole):
    """Modified version of code.interactiveconsole's interactive mode emulator; uses pygcurse IO. This gets stuck to a gameobjs.Player
    when the level starts so the player can have a nice interface to 'hack' the level with."""
    
    def __init__(self,mylocals,pygwindow):
        self.pygwindow=pygwindow
        super().__init__(mylocals)
        self.combuff=[]
        self.combuffpos=0
        self.backlog=""
        self.active=0
        
    def runcode(self, code):
        #We redirect stdout to... ourself, so that the code's output will go into the pygcurse window. 
        #We have a write() function, which is apparently all you actually need to be a stdio replacement. 
        oldstdout=sys.stdout
        sys.stdout=self
        super().runcode(code)
        #Then reset it. This isn't strictly necessary, but it's good form.
        sys.stdout=oldstdout
    
    def write(self, data):
        if self.active:
            self.pygwindow.pygprint(data)
        else:
            #If we're written to while we're not "active", we store the data in our
            #backlog instead; it'll be printed next time the user brings up the console
            self.backlog+=data
        
    def raw_input(self,prompt=""):
        #The way pygcurse does things necessitates some horrible nasty hax.
        #This will be made into a reference to the pyginput object pygwindow.input
        #uses to grab input, which allows us to change what the current input line is,
        #allowing such things as pasting and command history.
        #horrible, horrible nasty hax
        pyginput=None
        #Command buffer position -1 means we're not accessing the command buffer;
        self.combuffpos=-1
        def cback(pyginput=pyginput,injectionconsole=self):
            #*** BEGINNING OF HORRIFYING NASTY HAX ***
            #This will be called by pygcurse's input function every frame;
            #note that we've shoved some variables from the above scope into
            #our scope through default values; a sort of ghetto closure
            #First we check if we've already grabbed the PygcurseInput obj,
            #and if we haven't, do so
            if not pyginput:
                #The following code is made of satan.
                #Grabs the current stack frame... I think.
                frame = inspect.currentframe()
                try:
                    #This then looks at that frame, grabs the frame that called it,
                    #and grabs that frame's locals dict
                    caller_locals=frame.f_back.f_locals
                finally:
                    #Make sure we get rid of that reference to another frame, otherwise we can
                    #create a circular reference that confuses the GC and Guido will hunt
                    #us down and murder us in our sleep
                    del frame
                #please kill me
                #We then grab the pygcurse inputobj, which is the object that gets polled to handle keyboard input
                #for pygcurse (!?!?)
                pyginput=caller_locals["inputObj"]
                #We also want to get rid of THAT reference, just in case
                del caller_locals
                #So now we have our pyginput object. Hopefully Al will eventually
                #make it so pygcurse just passes this to us and we can delete all this godless shit.
                #*** END OF HORRIFYING NASTY HAX ***
                
            #Now that we have the pyginput object, we can do things to the current line being edited.
            #We can't use events for input, however, because pygcurse has already eaten them all.
            pressed=pygame.key.get_pressed()
            if pressed[pygame.locals.K_UP]:
                #Up key = go up one command in the history
                injectionconsole.combuffpos+=1
                #Don't want to access more command history than we have, obviously
                if injectionconsole.combuffpos >= len(injectionconsole.combuff):
                    injectionconsole.combuffpos=-1
                #Then we stuff that line of the command history into the pyginput buffer
                #assuming of course that we are actually accessing the buffer (-1 means "don't access the buffer")
                if injectionconsole.combuffpos != -1:
                    pyginput.buffer=list(injectionconsole.combuff[injectionconsole.combuffpos])
                else:
                    pyginput.buffer=list()
            elif pressed[pygame.locals.K_DOWN]:
                #Down key = go DOWN one command in the history
                #We don't want to go below -1, obviously
                injectionconsole.combuffpos-=1
                if injectionconsole.combuffpos < -1:
                    injectionconsole.combuffpos=len(injectionconsole.combuff)-1
                
                if injectionconsole.combuffpos != -1:
                    pyginput.buffer=list(injectionconsole.combuff[injectionconsole.combuffpos])
                else:
                    pyginput.buffer=list()
                    
            elif pressed[pygame.locals.K_v] and pygame.key.get_mods() == pygame.locals.KMOD_CTRL:
                #Finally, Ctrl-V means paste.
                #There would be pasting code here, but apparently there's no actual
                #way to do that in Python. Who knew! [rage]
                pass
                
        return self.pygwindow.input(prompt, fps=15, callbackfn=cback)
    
    def flush(self,**params):
        "Just here to keep python happy if we quit while we're stdio."
        pass
    
    def interact(self, banner=None):
        """Closely emulate the interactive Python console.

        The optional banner argument specifies the banner to print
        before the first interaction; by default it prints a banner
        similar to the one printed by the real Python interpreter,
        followed by the current class name in parentheses (so as not
        to confuse this with the real interpreter -- since it's so
        close!).


        I had to copy-n-edit this whole thing so there was a way to exit
        the shell; pygcurse input doesn't support ^Z, so I just made
        semicolon-enter the break command instead.
        """
        self.active=True
        #We immediately flush any text we were given while inactive to the screen
        if self.backlog:
            self.write(self.backlog)
            self.backlog=""
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "
        cprt = 'Type "help", "copyright", "credits" or "license" for more information.'
        if banner is None:
            self.write("Python %s on %s\n%s\n(%s)\nSemicolon then Enter to return to game" %
                       (sys.version, sys.platform, cprt,
                        self.__class__.__name__))
        elif banner:
            self.write("%s\n" % str(banner))
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                try:
                    line = self.raw_input(prompt)
                except EOFError:
                    self.write("\n")
                else:
                    if line==";":
                        #Semicolon is the "get out of this shell" command.
                        self.active=False
                        return
                    else:
                        #Before we process the line, stuff it in our command history list
                        self.combuff.insert(0,line)
                        more = self.push(line)
            except KeyboardInterrupt:
                self.write("\nKeyboardInterrupt\n")
                self.resetbuffer()
                more = 0

class Level:
    "A level of the game. Contains all the objects on a level."
    
    def __init__(self,name,objlist,plyobj,music=None):
        self.name=name
        self.objlist=objlist
        self.safe_objlist=[]
        self.consolefuncs=[]
        self.plyobj=plyobj
        self.plystartpos=plyobj.pos
        self.music=music
        
    def draw(self,pygcursewindow):
        "Call obj.draw on all objects in objlist and the plyobj."
        for theobj in self.safe_objlist:
            theobj.draw(pygcursewindow)
        for theobj in self.objlist:
            theobj.draw(pygcursewindow)
        self.plyobj.draw(pygcursewindow)
            
    def update(self,gamestate):
        "Call obj.update on all objects in oblist and the plyobj."
        
        #Gameobj updates do the stdout redirect trick too now
        oldstdout=sys.stdout
        sys.stdout=self.plyobj.injectionconsole
        
        for theobj in self.objlist:
            try:
                theobj.update(gamestate)
            except Exception:
                #Exceptions in a gameobj's code shouldn't crash the game; they should be printed to the injection console.
                #We only want the first two parts of exc_info printed; the traceback is meaningless
                #to the player. format_exception_only returns a list for god knows what reason, so that's why there's a [0]
                excstr=traceback.format_exception_only(*sys.exc_info()[:2])[0]
                #We also don't want to spam the console with the same exception over and over, so we check
                #to see if it's already there first
                if excstr not in self.plyobj.injectionconsole.backlog:
                    print(traceback.format_exception_only(*sys.exc_info()[:2])[0])
                #and an alert printed at the bottom of the screen
                gamestate.mainwin.putchars("Exception detected - See injection console for details.", 0, 29, fgcolor="red")
        
        for theobj in self.safe_objlist:
            try:
                theobj.update(gamestate)
            except Exception:
                excstr=traceback.format_exception_only(*sys.exc_info()[:2])[0]
                if excstr not in self.plyobj.injectionconsole.backlog:
                    print(excstr)
                gamestate.mainwin.putchars("Exception detected - See injection console for details.", 0, 29, fgcolor="red")
        
        self.plyobj.update(gamestate)
        
        sys.stdout=oldstdout
            
    def start(self,gamestate):
        "Called when this level becomes the current level, before its first update()"
        self.plyobj.injectionconsole=InjectionConsole({"objs":self.objlist, "funcs":self.consolefuncs},gamestate.mainwin)
        self.start_music(gamestate)
        gamestate.mainwin.fill()
        gamestate.mainwin.pygprint(self.name)
        gamestate.mainwin.update()
        pygcurse.waitforkeypress()
        
    def start_music(self,gamestate,loadedsave=False):
        #Music is either a string path to a music file or None
        try:
            #We do the loadedsave or music != lastmusic because we don't want to restart music that's already playing
            #but we DO want to play music if we just loaded a save, even if it's the music we were playing when we saved
            if loadedsave or (self.music and self.music != gamestate.lastmusic):
                if loadedsave:
                    music=gamestate.lastmusic
                else:
                    music=self.music
                pygame.mixer.music.load(os.path.join("music",music))
                pygame.mixer.music.play()
                gamestate.lastmusic=self.music
        except pygame.error:
            print("Failed to load music: ",self.music)
            
    def get_at(self,pos,include_player=False):
        """Return the gameobj at a given position, unless it's the player and include_player isn't true. 
        Returns None if there is no obj there. WILL screw up if there are multiple non-player objects in a square."""
        if include_player:
            if self.plyobj.pos == pos:
                return self.plyobj
            
        for theobj in self.objlist + self.safe_objlist:
            if theobj.int_pos() == pos:
                return theobj
            
    def remove_obj(self,theobj):
        """Remove the selected object from our lists. It's okay to pass this function objects that aren't actually
        in the level, though that's a stupid thing to do so don't do it anyway. WILL raise TypeError if you try
        to pass in the player, though."""
        try:
            self.objlist.remove(theobj)
        except ValueError:
            pass
        else:
            return
        
        try:
            self.safe_objlist.remove(theobj)
        except ValueError:
            pass
        else:
            return
        
        if theobj is self.plyobj:
            raise TypeError("Tried to remove_obj() the player. Don't do that.")

def wall_line(start,dir,count):
    """Return a list of Wall objects. The first will be in a position determined by the start parameter (as an x,y tuple)
    The rest will be in start + ( dir * 0 to count)"""
    startx, starty = start
    dirx, diry = dir
    retlist=[]
    for t in range(0,count):
        ofstx=startx+(dirx*t)
        ofsty=starty+(diry*t)
        retlist.append(gameobjs.Wall((ofstx,ofsty)))
    return retlist
        
def wall_room(start, size):
    "Return a list of Wall objects forming a square box. The upper left corner will be at start, and the walls will be size walls long"
    startx, starty = start
    retlist=(
             wall_line(start,(0,1),size)+
             wall_line(start,(1,0),size)+
             wall_line((startx+size,starty),(0,1),size)+
             wall_line((startx,starty+size),(1,0),size+1)
             )
    return retlist    

#LEVEL 0 : SLEEPING IN THE DIR(T)
#Player hopefully learns how to use dir() to find locals and attributes
#Either set the keypad's code to something known, set the gate to be open,
#move the gate, or just move the levelporter inside the room. A wealth of options.
level0=Level("LEVEL 0: SLEEPING IN THE DIR(T)",[],gameobjs.Player((10,10)),"lvl0.ogg")
gate=gameobjs.Gate((14,12))
level0.objlist= [gameobjs.Lvl1Keypad((13,10),gate),
                gameobjs.Levelporter((15,12),"Shortest")]
level0.safe_objlist=wall_room((9,9),5)
level0.safe_objlist.remove(level0.get_at((14,12)))
level0.objlist.append(gate)

#LEVEL 1 : A POCKET FULL OF .POS IES
#Player realizes objects can be moved by setting their pos
#Move the levelporter a few squares over to where the player is. Simple.
level1=Level("LEVEL 1: A POCKET FULL OF .POS IES",[],gameobjs.Player((11,11)))
level1.objlist=[gameobjs.Levelporter((16,11),"Shmokoban")]
level1.safe_objlist=wall_room((10,10),5) + wall_room((15,10),5)

#LEVEL 2 : A HARSH MISTRESS
#Player learns how boulders work; they move according to their gravity attribute
#Player also learns how pressureplates work; they trip whenever something is on them.
level2=Level("LEVEL 2: SHMOKOBAN / A HARSH MISTRESS",[],gameobjs.Player((15,5)))
level2.objlist=[gameobjs.Boulder((21,12))]
gate1=gameobjs.Gate((15,8))
gate2=gameobjs.Gate((15,22))
level2.safe_objlist=(
                     wall_room((13,22),4)+
                     wall_room((13,4),4)+
                     wall_line((14,8),(0,1),14)+
                     wall_line((16,8),(0,1),14)+
                     wall_line((20,8),(0,1),14)+
                     wall_line((22,8),(0,1),14)+
                     [gameobjs.Wall((21,7)),gameobjs.Wall((21,22))]+
                     [gameobjs.PressurePlate((21,8),gate1),gameobjs.PressurePlate((21,21),gate2)]+
                     [gameobjs.Levelporter((15,24),"Grabitty")]
                     )
level2.safe_objlist.remove(level2.get_at((15,8)))
level2.safe_objlist.remove(level2.get_at((15,22)))
level2.safe_objlist.extend((gate1,gate2))

#LEVEL 3 : YOU'RE DRAGGING ME... UP?
#Player learns how gravplates work
level3=Level("LEVEL 3: YOU'RE DRAGGING ME... UP?",[],gameobjs.Player((12,12)),"lvl3.ogg")
level3.objlist=[gameobjs.Boulder((21,12)),gameobjs.GravPad((21,22),(0,-1),3)]
gate1=gameobjs.Gate((15,13))
level3.safe_objlist=(
                     wall_line((20,8),(0,1),14)+
                     wall_line((22,8),(0,1),14)+
                     wall_room((10,10),5) + wall_room((15,10),5)+
                     [gameobjs.Wall((21,7))]+
                     [gameobjs.PressurePlate((21,18),gate1)]+
                     [gameobjs.Levelporter((19,13),"Dejavu")]
                     )
level3.safe_objlist.remove(level3.get_at((15,13)))
level3.safe_objlist.remove(level3.get_at((15,13))) #Gotta do it twice because we put TWO walls there
level3.safe_objlist.append(gate1)

#LEVEL 4 : ELECTRIC S(L)IDEWAYS
#Player gets some more practice with the gravpad, and learns it can force sideways as well.
level4=Level("LEVEL 4: ELECTRIC S(L)IDEWAYS",[],gameobjs.Player((12,12)))
level4.objlist=[gameobjs.GravPad((21,22),(0,-1),1), gameobjs.GravPad((21,9),(0,1),1)]
gate1=gameobjs.Gate((15,13))
gate2=gameobjs.Gate((17,15))
level4.safe_objlist=(
                     [gameobjs.Boulder((21,12))]+
                     wall_line((20,8),(0,1),14)+
                     wall_line((22,9),(0,1),13)+
                     wall_room((10,10),5) + wall_room((15,10),5) + wall_room((15,15),5)+
                     wall_line((21,7),(1,0),5)+
                     [gameobjs.PressurePlate((21,18),gate1), gameobjs.PressurePlate((25,8),gate2)]+
                     [gameobjs.Wall((26,8))]+
                     [gameobjs.Levelporter((19,18),"Missiles")]
                     )
level4.safe_objlist.remove(level4.get_at((15,13)))
level4.safe_objlist.remove(level4.get_at((15,13)))
level4.safe_objlist.remove(level4.get_at((17,15)))
level4.safe_objlist.remove(level4.get_at((17,15)))
level4.safe_objlist.extend((gate1,gate2))

#LEVEL 5 : THE LONG ARM
#Player learns how missiles and missile launchers work. He sees that the missile destroys the boulder, but not the wall,
#and figures out that he can re-target it to destroy the gate.
#Player is also introduced to attribute enforcers if he tries to get cute and move the gate.
level5=Level("LEVEL 5: THE LONG ARM",[],gameobjs.Player((11,15)),"lvl5.ogg")
gate1=gameobjs.Gate((30,15))
level5.objlist=[gameobjs.MissileLauncher((18,15),level5.plyobj),gameobjs.SafePressurePlate((21,18),gate1)]
level5.safe_objlist=(
                     wall_room((10,10),20)+
                     [gameobjs.Wall((12,15)), gameobjs.Wall((32,15)),gameobjs.Wall((31,14)),gameobjs.Wall((31,16))]+
                     [gameobjs.Boulder((14,12)), gameobjs.Wall((14,16))]+
                     [gameobjs.Levelporter((31,15),"Force"), gameobjs.AttributeEnforcer((35,12),{(gate1,"pos"):None})]
                     )
level5.safe_objlist.remove(level5.get_at((30,15)))
level5.safe_objlist.append(gate1)

#LEVEL 6 : SUPERIOR (EN)FORCE
#Player is formally introduced to attribute enforcers, and learns that they can be destroyed with missiles
#Also the first "action-y" level.
level6=Level("LEVEL 6: SUPERIOR (EN)FORCE",[],gameobjs.Player( (25, 10) ))
gravgen1=gameobjs.GravPad((52, 12),(0,1),3)
portergate=gameobjs.Gate((18, 10))
holdupgate=gameobjs.Gate((54, 9))
launcher=gameobjs.MissileLauncher((36, 18),level6.plyobj)
level6.objlist= [gravgen1,
                 gameobjs.Boulder((47, 2))]
level6.safe_objlist=[ 
                    gameobjs.Wall((24, 11)),
                    gameobjs.Wall((25, 11)),
                    gameobjs.Wall((26, 11)),
                    gameobjs.Wall((35, 11)),
                    gameobjs.Wall((36, 11)),
                    gameobjs.Wall((37, 11)),
                    gameobjs.Wall((45, 11)),
                    gameobjs.Wall((46, 11)),
                    gameobjs.Wall((47, 11)),
                    gameobjs.Wall((49, 11)),
                    gameobjs.Wall((50, 11)),
                    gameobjs.Wall((51, 11)),
                    gameobjs.Wall((52, 11)),
                    gameobjs.Wall((49, 7)),
                    gameobjs.Wall((51, 7)),
                    gameobjs.Wall((50, 7)),
                    gameobjs.Wall((52, 7)),
                    gameobjs.Wall((53, 7)),
                    gameobjs.Wall((53, 8)),
                    gameobjs.Wall((53, 9)),
                    gameobjs.Wall((53, 10)),
                    gameobjs.Wall((53, 11)),
                    gameobjs.Wall((55, 7)),
                    gameobjs.Wall((55, 9)),
                    gameobjs.Wall((55, 10)),
                    gameobjs.Wall((55, 11)),
                    gameobjs.Wall((54, 11)),
                    gameobjs.Wall((54, 7)),
                    gameobjs.PressurePlate((52, 8),holdupgate),
                    gameobjs.Wall((22, 11)),
                    gameobjs.Wall((22, 9)),
                    gameobjs.Wall((21, 9)),
                    gameobjs.Wall((20, 9)),
                    gameobjs.Wall((19, 9)),
                    gameobjs.Wall((18, 9)),
                    gameobjs.Wall((21, 11)),
                    gameobjs.Wall((20, 11)),
                    gameobjs.Wall((19, 11)),
                    gameobjs.Wall((18, 11)),
                    gameobjs.Wall((17, 9)),
                    gameobjs.Wall((17, 11)),
                    gameobjs.Wall((17, 8)),
                    gameobjs.Wall((14, 11)),
                    gameobjs.Wall((16, 11)),
                    gameobjs.Wall((15, 11)),
                    gameobjs.Wall((13, 11)),
                    gameobjs.Wall((13, 10)),
                    gameobjs.Wall((13, 9)),
                    gameobjs.Wall((13, 8)),
                    gameobjs.Wall((13, 7)),
                    gameobjs.Wall((14, 7)),
                    gameobjs.Wall((16, 7)),
                    gameobjs.Wall((15, 7)),
                    gameobjs.Wall((17, 7)),
                    gameobjs.Levelporter((14, 8),'DroneSpiral'),
                    gameobjs.Boulder((54, 8)),
                    gameobjs.Wall((55, 8)),
                    gameobjs.PressurePlate((54, 10),portergate),
                    gameobjs.AttributeEnforcer((46, 4),{(gravgen1,"gravadj"):None,(gravgen1,"pos"):None}),
                    gameobjs.Boulder((52, 9)),
                    gameobjs.Wall((26, 10)),
                    portergate,
                    holdupgate,
                    launcher]
level6.safe_objlist.extend(wall_room((23,7),25))
level6.safe_objlist.extend(wall_room((41, 0),7))
level6.safe_objlist.extend(wall_line((51,8),(0,1),3))
level6.remove_obj(level6.get_at((23,10)))
level6.remove_obj(level6.get_at((48,9)))
level6.remove_obj(level6.get_at((42,7)))
level6.remove_obj(level6.get_at((42,7)))

#LEVEL 7 : SPIRAL POWER
#Player is introduced to drones/droneupload, programming a drone to escape a simple spiral "maze".
level7=Level("LEVEL 7: SPIRAL POWER/GETTIN FUNC-Y (check the 'funcs' list)",[],gameobjs.Player( (25, 15) ), "lvl7.ogg")
gate1=gameobjs.Gate((27, 3))
gate2=gameobjs.Gate((30, 15))
drone=gameobjs.Drone((35, 7))
level7.objlist=[]
level7.consolefuncs=[drone.upload]
level7.safe_objlist=[ gameobjs.Wall((35, 6)),
                   gameobjs.Wall((34, 6)),
                   gameobjs.Wall((34, 7)),
                   gameobjs.Wall((34, 8)),
                   gameobjs.Wall((35, 8)),
                   gameobjs.Wall((36, 8)),
                   gameobjs.Wall((37, 8)),
                   gameobjs.Wall((37, 7)),
                   gameobjs.Wall((37, 6)),
                   gameobjs.Wall((37, 5)),
                   gameobjs.Wall((37, 4)),
                   gameobjs.Wall((36, 4)),
                   gameobjs.Wall((34, 4)),
                   gameobjs.Wall((35, 4)),
                   gameobjs.Wall((33, 4)),
                   gameobjs.Wall((32, 4)),
                   gameobjs.Wall((32, 5)),
                   gameobjs.Wall((32, 6)),
                   gameobjs.Wall((32, 7)),
                   gameobjs.Wall((32, 8)),
                   gameobjs.Wall((32, 9)),
                   gameobjs.Wall((32, 10)),
                   gameobjs.Wall((33, 10)),
                   gameobjs.Wall((34, 10)),
                   gameobjs.Wall((35, 10)),
                   gameobjs.Wall((37, 10)),
                   gameobjs.Wall((36, 10)),
                   gameobjs.Wall((38, 10)),
                   gameobjs.Wall((39, 10)),
                   gameobjs.Wall((39, 9)),
                   gameobjs.Wall((39, 8)),
                   gameobjs.Wall((39, 7)),
                   gameobjs.Wall((39, 6)),
                   gameobjs.Wall((39, 5)),
                   gameobjs.Wall((39, 4)),
                   gameobjs.Wall((39, 3)),
                   gameobjs.Wall((39, 2)),
                   gameobjs.Wall((38, 2)),
                   gameobjs.Wall((36, 2)),
                   gameobjs.Wall((37, 2)),
                   gameobjs.Wall((35, 2)),
                   gameobjs.Wall((34, 2)),
                   gameobjs.Wall((32, 2)),
                   gameobjs.Wall((33, 2)),
                   gameobjs.Wall((31, 2)),
                   gameobjs.Wall((31, 4)),
                   gameobjs.Wall((30, 4)),
                   gameobjs.Wall((30, 2)),
                   gameobjs.Wall((29, 2)),
                   gameobjs.Wall((29, 4)),
                   gameobjs.Wall((28, 4)),
                   gameobjs.Wall((28, 2)),
                   gameobjs.Wall((27, 2)),
                   gameobjs.Wall((27, 4)),
                   gate1,
                   gameobjs.PressurePlate((28, 3),gate1),
                   gameobjs.Wall((26, 2)),
                   gameobjs.Wall((26, 4)),
                   gameobjs.Wall((25, 3)),
                   gameobjs.Wall((25, 14)),
                   gameobjs.Wall((26, 14)),
                   gameobjs.Wall((27, 14)),
                   gameobjs.Wall((28, 14)),
                   gameobjs.Wall((29, 14)),
                   gameobjs.Wall((30, 14)),
                   gameobjs.Wall((31, 14)),
                   gameobjs.Wall((32, 14)),
                   gameobjs.Wall((33, 14)),
                   gameobjs.Wall((34, 14)),
                   gameobjs.Wall((35, 14)),
                   gameobjs.Wall((37, 14)),
                   gameobjs.Wall((36, 14)),
                   gameobjs.Wall((38, 14)),
                   gameobjs.Wall((39, 14)),
                   gameobjs.Wall((39, 16)),
                   gameobjs.Wall((38, 16)),
                   gameobjs.Wall((37, 16)),
                   gameobjs.Wall((36, 16)),
                   gameobjs.Wall((35, 16)),
                   gameobjs.Wall((34, 16)),
                   gameobjs.Wall((33, 16)),
                   gameobjs.Wall((31, 16)),
                   gameobjs.Wall((32, 16)),
                   gameobjs.Wall((30, 16)),
                   gameobjs.Wall((29, 16)),
                   gameobjs.Wall((28, 16)),
                   gameobjs.Wall((27, 16)),
                   gameobjs.Wall((26, 16)),
                   gameobjs.Wall((25, 16)),
                   gameobjs.Wall((24, 15)),
                   gameobjs.Wall((40, 15)),
                   gameobjs.Levelporter((39, 15),"END"),
                   gate2,
                   gameobjs.PressurePlate((26, 3),gate2),
                   drone]

levels={
        "Start" : level0,
        "Shortest" : level1,
        "Shmokoban" : level2,
        "Grabitty" : level3,
        "Dejavu" : level4,
        "Missiles" : level5,
        "Force" : level6,
        "DroneSpiral" : level7
        }
