"""
INJECTION
By Schilcote
For Pyweek 19
Main Program Source File
Monday, Oct. 6, 2014
"""

import pygame
from pygame.locals import *
import pygcurse
import levels
import pickle
import os
import time
import imp
import random
import sys
import socket
import hashlib
import io
import zipfile

#Our actual version number for autoupdate purposes will be prefixed with our platform and/or freeze status
_version="11-11"
#Set this to true before distributing. Disables loading of saves and auto-updates if false. 
_distributable=True
#This can be set to true by the arg parser to disable automatic updates.
_disable_update=False
#This is the address/URL of the update server
_update_addr="iju.404.mn"
#This is the font the game will use for rendering; should be a Pygame-compatible, unicode-supporting font.
_font_path="fsex2.ttf"

class GameState:
    "Holds the game's global state. Also holds saving/loading functions."
    
    def __init__(self,mainwin):
        """Screen: Pygame surf. obj from display.set_mode
        Mainwin: PygcurseWindow encompassing the entire Pygame screen"""

        self.mainwin=mainwin
        self.current_level=None
        self.current_level_name=""
        self.controls_locked=False
        self.lastmusic=""
        
    def change_level(self,level):
        "Called to change the current level and do the necessary things to that level before actually playing it."
        pump_events(self,True) #Flush the event queue so we don't just flash the level name because there's keypresses in the pipeline
        if level=="END":
            self.mainwin.fill()
            self.mainwin.cursor=(0,0)
            self.mainwin.pygprint("""
            YOU'RE WINNER
            
            Oh dear, we seem to have run out of levels-
            for now! This game is still in development;
            new levels are added every so often,  and 
            the game automatically updates itself, so 
            check back in a week or so! 
            
            Drop me a line at
            toastengineer@gmail.com
            if you enjoyed the game, and would 
            like to see it finished.
            
            Thanks for playing!
                                """)
            self.mainwin.update()
            pump_events(self,True)
            time.sleep(1)
            pump_events(self,True)
            pygcurse.waitforkeypress()
            try:
                os.remove("save.sav")
            except:
                pass
            exit()
        self.current_level=levels.levels[level]
        self.curent_level_name=level
        self.current_level.start(self)
        time.sleep(1)
        self.save()
        
    def reset_level(self):
        "Confirm that the user actually wants to reset; if so, reload the levels module (to reset the level) and refresh the current level"
        self.mainwin.fill()
        self.mainwin.putchars("Are you SURE you want to reset?",5,0)
        self.mainwin.putchars("Type 'yes' to confirm; anything else to cancel.",5,1)
        yesno=self.mainwin.input(">")
        if yesno.lower()=="yes":
            imp.reload(levels)
            self.change_level(self.curent_level_name)
        else:
            return
        
    def update(self):
        "Tell the level to update, then the pygcurse window."
        self.current_level.update(self)
        self.current_level.draw(self.mainwin)
        self.mainwin.update()
        
        
    def save(self):
        "Save the game."
        #Eh, fuck it, just pickle gamestate. What could go wrong?
        pickle.dump(self,open("save.sav","wb"),protocol=pickle.HIGHEST_PROTOCOL)
        
    def save_and_quit(self):
        "Save the game and terminate execution."
        self.save()
        exit()
        
def autoupdate():
    "Do the auto-update procedure. Quits and restarts the program if an update is necessary."
    
    #If we're an in-progress version, we damn well better not overwrite ourself
    if not _distributable:
        return
    
    #If autoupdates have been turned off at the command line, well...
    if _disable_update:
        return
    
    #This'll evaluate true if we're a cx_freeze frozen distro
    if getattr(sys, 'frozen', False):
        #If we're frozen, OS matters (maybe...). So our plat_id is just the return value of sys.platform
        #This'll be win32 for windows, linux2 for linux, darwin for MacOS
        plat_id=sys.platform
    else:
        #If we're not frozen, we must be a source distro
        #and there's only one version of the source; OS is irrelevant
        plat_id="src"
    
    #Otherwise, begin the auto-update procedure.
    #iju.404.mn is set to point to the autoupdate server.
    with socket.create_connection((_update_addr,7363)) as connection:
        connection.send(bytes(plat_id+_version+"\n","ASCII"))
        filehash=connection.recv(16) #MD5 hashes are always 16 bytes
        connection.send(b"\x06")
        file=bytes()
        while True:
            stuff=connection.recv((2**16)*10)
            if stuff:
                file+=stuff
            else:
                break
    #It'll fail here if we're at the latest version (because the last read gives us nothing, which obviously doesn't hash the same as
    #our archive
    if not filehash == hashlib.md5(file).digest():
        return
    archive=io.BytesIO(file)
    zipdata=zipfile.ZipFile(archive)
    zipdata.extractall(os.path.curdir)
    args = sys.argv[:]
    args.insert(0, sys.executable)
    os.execv(sys.executable, args)
    
def do_cmdline_args():
    "Do our own goddamn command line processing. Screw argparse. Ugh."
    argv=sys.argv
    for k, arg in enumerate(argv):
        
        if arg=="-force_backend":
            print("Forcing Pygame backend to",argv[k+1])
            os.environ["SDL_VIDEODRIVER"]=argv[k+1]
        elif arg=="-disable_update":
            print("Disabing automatic updates for this session.")
            _disable_update=True
        elif arg=="-update_addr":
            print("Using alternate update server at",argv[k+1])
            _update_addr=argv[k+1]
        elif arg=="-font":
            print("Using alternative font",argv[k+1])
            _font_path=argv[k+1]
        elif arg=="-help" or arg=="-h" or arg=="-?" or arg=="/?" or arg=="-v":
            print("INJECTION Version",_version)
            print("By Schilcote")
            print("")
            print("-help : Display this text")
            print("-force_backend [backend] : Force the game to to use a specific rendering backend. Try directx on Windows to allow FRAPS to work.")
            print("-disable_update : Disable the auto-update functionality.")
            print("-update_addr [addr]: Grab auto-updates from the given server.")
            print("-font [path]: Force INJECTION to use the specified font. Should be a full path to a font file.")
            exit()
        
def game_init():
    "Performs primary initialization. Returns a GameState object, which'll be this sessions global state manager."
    do_cmdline_args()
    
    if sys.version_info.major != 3:
        print("INJECTION is designed to run with Python 3 or greater. You are running",sys.version[:5])
        time.sleep(3)
    
    pygame.init()
    mainwin=pygcurse.PygcurseWindow(80, 30, "INJECTION")
    mainwin.autoupdate=False #We don't want auto-blit; otherwise it redraws every time an individual object draws and flickers
    
    try:
        mainwin.font=pygame.font.Font(_font_path, 18)
        pass
    except:
        print("Could not load font. Reverting to default. Note that INJECTION will not work properly without a unicode font.")
    
    pygame.key.set_repeat(50,100)
    
    try:
        if not _distributable:
            raise Exception
        gamestate=pickle.load(open("save.sav","rb"))
        gamestate.mainwin=mainwin
        #We've got to start the music here when we load a save; the level
        #init doesn't get called.
        gamestate.current_level.start_music(gamestate,True)
        #We ALSO need to reset the injection console's version of mainwin
        gamestate.current_level.plyobj.injectionconsole.pygwindow=mainwin
    except Exception:
        gamestate=GameState(mainwin)

    return gamestate

def display_attract_screen(mainwin):
    "Displays the 'attract mode' screen with title, basic instructions. Blocks until user presses a key."
    mainwin.pygprint(
                     """
                     INJECTION
                     
                     PRESS ANY KEY TO CONTINUE
                     
                     ARROWKEYS to move
                     Walk into objects to interact with them
                     ; to inject
                     R to reset the level
                     ESCAPE saves and quits."""
                     )
    mainwin.update()
    pygcurse.waitforkeypress()
    
def pump_events(gamestate,flush=False):
    """This actually mostly exists just to empty out the Pygame event queue; without it it'll fill with events the
     plyobj doesn't care about. It also handles the game being quit through the OS."""
    for event in pygame.event.get():
        if event.type==QUIT:
            gamestate.save_and_quit()
        if event.type==KEYDOWN:
            #gameobjs.Player handles keydown events, so we have to put THEM back in the queue 
            #Yes, this is very very stupid. I'm sorry.
            #We don't worry about the whole "don't edit an iterable you're iterating over" thing because
            #pygame.event.get just hands back a list of how the queue was when it was called.
            #If flush is true, though, then we don't because we're trying to empty the queue
            #@TODO Fix this so that there's a custom "moveplayerobj" event instead.
            if not flush:
                pygame.event.post(event)
                
def do_glitch_effect(surface):
    buffer=memoryview(surface.get_buffer())
    size=len(buffer)
    for px in range(0,(size-3*4)):
        if not random.randint(0,1):
            buffer[px]=buffer[px+(random.randint(-3,3)*4)]
    buffer.release()
    del(buffer)
    surface.unlock()
    pygame.display.flip()


if __name__ == '__main__':
    #First, check for/install updates. This'll raise various types of exceptions if it fails; we just silence them and carry on.
    try:
        autoupdate()
    except Exception:
        pass
    gamestate=game_init()
    display_attract_screen(gamestate.mainwin)
    #gamestate.current_level will be None if we've just started; if it's not, we loaded a save.
    if not gamestate.current_level:
        gamestate.change_level("Start")
    clock=pygame.time.Clock()

    while True:
        #Blank the screen
        gamestate.mainwin.fill()
        #The gamestate obj. relays the update order to the level and Pygcurse
        gamestate.update()
        #Do global-level event stuff
        pump_events(gamestate)
        #Only burn as much CPU as we need
        clock.tick(15)
        #Repeat until smooth and without clumps
        